package virtual::xen;

use common;
use run_program;

my $xen_config_dir = "/etc/xen";
my $xen_guests_dir = $xen_config_dir . '/guests';
my $xend_config = $xen_config_dir . '/xend-config.sxp';

sub ensure_is_installed {
    my ($_class, $in) = @_;

    $in->wait_message(N("Checking availability of %s components", "Xen"));

    $in->do_pkgs->is_installed('lilo') &&
      !$in->do_pkgs->ensure_is_installed("mbootpack", "/usr/bin/mbootpack") and return;
    $in->do_pkgs->ensure_is_installed("xen", "/boot/xen.gz") or return;

    require services;
    # the unix server is required by libvirt when run as root
    if (!any { /^\s*\(xend-unix-server\s+yes\)/ } cat_($xend_config)) {
        my $has_unix_server;
        substInFile {
            s/^\s*#*\s*\(xend-unix-server .*/(xend-unix-server yes)/ and $has_unix_server = 1;
            $_ .= "(xend-unix-server yes)\n" if eof && !$has_unix_server;
        } $xend_config;
        is_available() and services::restart_or_start("xend");
    }

    foreach ("xend", "xendomains") {
        services::start_service_on_boot($_);
        services::start_not_running_service($_) if is_available();
    }

    1;
}

sub is_available {
    any { /\[xenbus\]$/ } `ps x`;
}

sub get_not_available_message {
    N("You must reboot on a Xen kernel to be able to run virtualized guests.");
}

sub get_guest_domain_settings {
    my ($self) = @_;
    [
        {
            label => N("Guest domain name"),
            val => \$self->{guest}{name},
        },
    ];
}

sub check_guest_domain_settings {
    my ($self) = @_;
    if (!$self->{guest}{name}) {
        $self->{guest}{error}{message} = N("Please specify a guest domain name");
        $self->{guest}{error}{field} = \$self->{guest}{name};
        return 0;
    }
    return 1;
}

sub get_guest_filesystem_settings {
    my ($self) = @_;
    [
        {
            label => N("Guest filename"),
            type => 'file',
            val => \$self->{guest}{filename},
        },
        {
            label => N("Guest size (MB)"),
            type => 'range',
            min => 400,
            max => 20000,
            val => \$self->{guest}{size},
        },
        {
            label => N("Installer path"),
            type => 'file',
            want_a_dir => 1,
            val => \$self->{guest}{installer_path},
        },
    ];
}

sub check_guest_filesystem_settings {
    my ($self) = @_;
    if (!$self->{guest}{filename}) {
        $self->{guest}{error}{message} = N("Please specify a guest filename");
        $self->{guest}{error}{field} = \$self->{guest}{filename};
        return 0;
    }
    if (!$self->{guest}{installer_path}) {
        $self->{guest}{error}{message} = N("Please specify the installer path");
        $self->{guest}{error}{field} = \$self->{guest}{installer_path};
        return 0;
    }
    return 1;
}

sub create_guest_file {
    my ($self) = @_;
    mkdir_p(dirname($self->{guest}{filename}));
    run_program::run('dd', 'if=/dev/zero', 'of=' . $self->{guest}{filename}, 'bs=1M', 'count=1', 'seek=' . $self->{guest}{size}) or return;
    run_program::run('mke2fs', '-F', '-j', $self->{guest}{filename}) or return;
    1;
}

sub get_drakx_in_chroot {
    my ($repository) = @_;
    my $drakx_in_chroot = $repository . '/misc/drakx-in-chroot';
    my $remote_repository = $repository =~ m!^(ftp|http)://! && $1;
    if ($remote_repository) {
        my $local_drakx_in_chroot = '/tmp/drakx-in-chroot-real';
        mkdir_p(dirname($local_drakx_in_chroot));
        run_program::run('curl', '--silent', '-o', $local_drakx_in_chroot, $drakx_in_chroot)
          or die "unable to get drakx-in-chroot from remote repository";
        $drakx_in_chroot = $local_drakx_in_chroot;
    }
    $drakx_in_chroot;
}

sub run_guest_installer {
    my ($self, $in) = @_;

    $in->do_pkgs->ensure_are_installed([ 'xorg-x11-Xnest', 'curl' ]) or return;

    require File::Temp;
    (undef, my $defcfg) = File::Temp::tempfile(undef, OPEN => 0);
    output($defcfg, <<EOF);
\$o = {
  'default_packages' => [ 'kernel-xen' ],
};
EOF

    my $mnt = File::Temp::tempdir();
    run_program::run('mount', '-o', 'loop', $self->{guest}{filename}, $mnt) or return;
    my $drakx_in_chroot = get_drakx_in_chroot($self->{guest}{installer_path});
    run_program::raw({ timeout => 'never' }, 'perl', $drakx_in_chroot, $self->{guest}{installer_path}, $mnt, '--defcfg', $defcfg) or return;
    run_program::run('umount', $mnt);

    1;
}

sub configure_guest {
    my ($self) = @_;
    output_p($xen_guests_dir . '/' . $self->{guest}{name}, <<EOF);
kernel = "/boot/vmlinuz-2.6-xen"
memory = 64
name = "$self->{guest}{name}"
disk = [ 'file:$self->{guest}{filename},hda1,w' ]
root = "/dev/hda1 ro"
hostname = "$self->{guest}{name}"
EOF
}

sub install_guest {
    my ($self, $in) = @_;
    if (!$self->create_guest_file) {
        $in->ask_warn(N("Error"), N("Unable to create guest file"));
        return;
    }
    if (!$self->run_guest_installer($in)) {
        $in->ask_warn(N("Error"), N("Unable to install guest"));
        return;
    }
    $self->configure_guest;
    1;
}

sub list_configured_guests {
    sort { $a->{name} cmp $b->{name} } map {
        virtual::guest::xen->new_from_file($xen_guests_dir . '/' . $_);
    } all($xen_guests_dir);
}

sub list_running_guests {
    my @guests = run_program::get_stdout('xm', 'list');
    shift @guests; #- skip header
    map {
        my ($name, $id) = /^(\S+)\s*(\S+)/;
        if_($id != 0, virtual::guest::xen->new($name));
    } @guests;
}

1;


package virtual::guest::xen;

use MDK::Common;
use run_program;

sub new {
    my ($class, $name) = @_;
    bless {
        name => $name,
    }, $class;
}

sub new_from_file {
    my ($class, $file) = @_;
    my $name = first(cat_($file) =~ m|^\s*name\s*=\s*"(.*)"$|m);
    bless {
        name => $name || basename($file),
        file => $file,
    }, $class;
}

sub start {
    my ($self) = @_;
    run_program::run('xm', 'create', $self->{file});
}

sub is_running {
    my ($self) = @_;
    system("xm >/dev/null 2>/dev/null list $self->{name}") == 0;
}

sub stop {
    my ($self) = @_;
    run_program::run('xm', 'shutdown', $self->{name});
}

sub kill {
    my ($self) = @_;
    run_program::run('xm', 'destroy', $self->{name});
}

sub get_console_command {
    my ($self) = @_;
    'xm', 'console', $self->{name};
}

1;
